# lexer.py
# Author: Thomas MINIER - MIT License 2019
import re
from pyparsing import CaselessKeyword, Keyword, LineEnd, Literal, MatchFirst, OneOrMore, Optional, Group, Regex, Word, ZeroOrMore

# ----- General terms ------

uriref = r'(<([^:]+:[^\s"<>]+)>|(([A-Za-z0-9]|-)+):([A-Za-z0-9]+))'
literal = r'"([^"\\]*(?:\\.[^"\\]*)*)"'
litinfo = r'(?:@([a-zA-Z]+(?:-[a-zA-Z0-9]+)*)|\^\^' + uriref  + r')?'

r_line = re.compile(r'([^\r\n]*)(?:\r\n|\r|\n)')
r_wspace = re.compile(r'[ \t]*')
r_wspaces = re.compile(r'[ \t]+')
r_tail = re.compile(r'[ \t]*\.[ \t]*(#.*)?')
r_uriref = re.compile(uriref)
r_nodeid = re.compile(r'_:([A-Za-z0-9]*)')
r_literal = re.compile(literal + litinfo)
r_variable = re.compile(r'\?([A-Za-z0-9]+)')
r_prefix = re.compile(r'([A-Za-z0-9]|-)+')

# a suppressed comma (',')
comma = Literal(',').suppress()

# A Turtle Prefix
prefixName = Regex(r_prefix)

# The special 'none' keyword (a shorthand notation for ottr:None)
ottrNone = Keyword('none')

# a RDF IRI
iri = Regex(r_uriref)

# a RDF Blank Node
bnode = Regex(r_nodeid)

# a SPARQL variable
variable = Regex(r_variable)

# a RDF Literal
literal = Regex(r_literal)

# An IRI or a Variable
iriOrVariable = MatchFirst([iri, variable])

# Any valid RDF terms
rdfTerm = MatchFirst([ottrNone, iri, literal, bnode, variable])

# Any valid RDF terms, excluding SPARQL variables
rdfTermNoVars = MatchFirst([ottrNone, iri, literal, bnode])

# ----- stOTTR language rules ------

# A template parameter definition, with optional type and nonblank
# Examples: "?iri", "xsd:string ?literal", "! otrr:IRI ?iri" or "?iri = ex:Ann"
param = Group(
            Optional(Keyword('!')).setResultsName('nonblank') +
            Optional(Keyword('?')).setResultsName('optional') +
            Optional(iri).setResultsName('type') +
            variable.setResultsName('value') +
            Optional(Keyword('=') + rdfTermNoVars.setResultsName('default'))
        ).setResultsName('parameter') + Optional(',').suppress()

# A list of template parameters
paramList = Group(
                Literal('[').suppress() +
                ZeroOrMore(param) +
                Literal(']').suppress()
            )

# An instance of a template which may contains variables
# like ottr:Triple (_:person, rdf:type, ?person)
instanceWithVars = Group(
                    iri.setResultsName('name') +
                    Literal('(').suppress() +
                    OneOrMore(rdfTerm + Optional(comma).suppress()).setResultsName('arguments') +
                    Literal(')').suppress()
                )
# An instance of a template which cannot contains variables
# like ottr:Triple (_:person, rdf:type, ?person)
instanceNoVars = Group(
                    iri.setResultsName('name') +
                    Literal('(').suppress() +
                    OneOrMore(rdfTermNoVars + Optional(comma).suppress()).setResultsName('arguments') +
                    Literal(')').suppress()
                )

# A stOTTR prefix declaration
prefixDeclaration = Group(
                CaselessKeyword("@prefix").suppress() +
                prefixName.setResultsName('name') +
                Literal(':').suppress() +
                iri.setResultsName('value') +
                Literal('.').suppress()
            )

# A stOTTR template
ottrTemplate = Group(
                iri.setResultsName('name') +
                paramList.setResultsName('parameters') +
                Literal('::').suppress() +
                Literal('{').suppress() +
                ZeroOrMore(instanceWithVars + Optional(',').suppress()).setResultsName('instances') +
                Literal('}').suppress() + Literal('.').suppress()
            )

# Several stOTTR templates
ottrRoot = ZeroOrMore(prefixDeclaration + LineEnd().suppress()).setResultsName('prefixes') + OneOrMore(ottrTemplate + LineEnd().suppress()).setResultsName('templates')

# Several concrete stOTTR instances (with no variables allowed)
ottrRootInstances = ZeroOrMore(prefixDeclaration + LineEnd().suppress()).setResultsName('prefixes') + OneOrMore(instanceNoVars + Keyword('.').suppress() + Optional(LineEnd()).suppress()).setResultsName('instances')


def lex_templates_stottr(text):
    """Run the lexer on a set of stOTTR template defintions"""
    return ottrRoot.parseString(text)

def lex_instances_stottr(text):
    """Run the lexer on a set of stOTTR instances"""
    return ottrRootInstances.parseString(text)



# utils.py
# Author: Thomas MINIER - MIT License 2019
from rdflib.namespace import Namespace

# OTTR namespace
OTTR = Namespace('http://ns.ottr.xyz/0.4/')
OTTR_RDF = Namespace('http://tpl.ottr.xyz/rdf/0.1/')
OTTR_RDFS = Namespace('http://tpl.ottr.xyz/rdfs/0.1/')

# OTTR special URIs
OTTR_NONE = OTTR['None']
OTTR_IRI = OTTR.IRI

# URI for OTTR base templates
OTTR_TRIPLE_URI = OTTR.Triple



# rdfs.py
# Author: Thomas MINIER - MIT License 2019

RDFS_TEMPLATES = """
@prefix o-rdfs: <http://tpl.ottr.xyz/rdfs/0.1/> .

o-rdfs:Label[ ottr:IRI ?iri, ?label ] :: {
    ottr:Triple(?iri, rdfs:label, ?label)
} .

o-rdfs:ResourceDescription[ ottr:IRI ?iri, ? ?label, ? ?comment, ? ?seeAlso, ? ?isDefinedBy ] :: {
    o-rdfs:Label(?iri, ?label),
    ottr:Triple(?iri, rdfs:comment, ?comment),
    ottr:Triple(?iri, rdfs:seeAlso, ?seeAlso),
    ottr:Triple(?iri, rdfs:isDefinedBy, ?isDefinedBy)
} .

o-rdfs:TypedResourceDescription[ ottr:IRI ?iri, ottr:IRI ?type, ? ?label, ? ?comment, ? ?seeAlso, ? ?isDefinedBy ] :: {
    ottr:Triple(?iri, rdf:type, ?type),
    o-rdfs:ResourceDescription(?iri, ?label, ?comment, ?seeAlso, ?isDefinedBy)    
} .
"""



# template.py
# Author: Thomas MINIER - MIT License 2019
from abc import ABC, abstractmethod
from rdflib import BNode, Literal, URIRef, Variable
from rdflib.namespace import RDFS
from ottr.base.utils import OTTR_IRI, OTTR_NONE

class TemplateParameter(object):
    """
        The parameter of an OTTR Template.

        Arguments:
            - name ``str``: parameter's name.
            - param_type :class`rdflib.term.URIRef`: parameter's type.
            - optional ``bool``: if the parameter is optional or not.
            - nonblank ``bool``: if the parameter accepts blank node or not.
            - default :class`rdflib.term.Identifier`: (optional) the parameter's default value if it is set to ottr:None.
    """

    def __init__(self, name, param_type, optional, nonblank, default=None):
        super(TemplateParameter, self).__init__()
        self._name = name
        self._param_type = param_type
        self._optional = optional
        self._nonblank = nonblank
        self._default = default
        # parameter with a default value are automatically set to optional
        if default is not None:
            self._optional = True

    @property
    def name(self):
        """Get the parameter name, i.e., a SPARQL variable"""
        return self._name

    def __str__(self):
        res = "{} {}".format(self._param_type.n3(), self._name.n3())
        if self._optional:
            res = "? " + res
        if self._nonblank:
            res = "! " + res
        return res

    def __repr__(self):
        return self.__str__()

    def validate(self, value):
        """
            Assert that an argument can be used for this parameter, i.e., its value is compatible with the parameter definition (type, optional, non blank, etc).

            Argument:
                - value :class`rdflib.term.Identifier`: Argument value
            Returns:
                A tuple ``(is_valid, value, error_reason)`` where:
                    - ``is_valid`` is a boolean which indicates the outcome of the validation.
                    - ``value`` is the RDF value to use for the argument. Set to ``None`` if the validation was not successfull.
                    - ``error_reason`` is an error message that indicates why the validation has failed. Set to ``None`` if the validation was successfull.
        """
        # assert that the argument's type is correct
        if self._param_type != RDFS.Resource:
            # validate uris
            if self._param_type == OTTR_IRI and (type(value) == Literal or type(value) == Variable):
                return False, None, "expected an IRI but instead got a {}".format(type(value))
            elif type(value) == Literal and value.datatype != self._param_type:
                return False, None, "expected a Literal with datatype {} but instead got {}".format(self._param_type.n3(), value.datatype)
        # assert that a non-optional parameter cannot be bound to ottr:None
        if not self._optional and value == OTTR_NONE:
            return False, None, 'this parameter is not optional, so it cannot be bound to ottr:none.'
        # assert that a non blank parameter is not bound to a blank node
        if self._nonblank and type(value) == BNode:
            return False, None, 'this parameter is non blank, so it cannot be bound to a blank node.'
        # if the value is None but has a default value, use the default value
        if value == OTTR_NONE and self._default is not None:
            return True, self._default, None
        # otherwise, everything is fine :-)
        return True, value, None

class AbstractTemplate(ABC):
    """An abstract OTTR Template."""

    def __init__(self, name):
        super(AbstractTemplate, self).__init__()
        self._name = name
        self._parameters = dict()

    @property
    def name(self):
        return self._name

    @abstractmethod
    def expand(self, arguments, all_templates, as_nt=False):
        """Returns a generator that expands the template"""
        pass

    def add_parameter(self, name, position, param_type=RDFS.Resource, optional=False, nonblank=False, default=None):
        """
            Register a new template parameter.

            Arguments:
                - name ``str``: parameter's name.
                - position ``int``: parameter's position in the template definition.
                - param_type :class`rdflib.term.URIRef`: (optional) parameter's type. Default is rdfs:Resource.
                - optional ``bool``: (optional) if the parameter is optional or not. Default is `False`.
                - nonblank ``bool``: (optional) if the parameter accepts blank node or not. Default is `False`.
                - default :class`rdflib.term.Identifier`: (optional) the parameter's default value if it is set to ottr:None.
        """
        if position not in self._parameters:
            self._parameters[position] = TemplateParameter(name, param_type, optional, nonblank, default=default)

    def format_arguments(self, arguments):
        """Format a list of execution arguments, i.e., a list of tuple (position, value), so that they can be used for template expansion"""
        args = dict()
        for position, value in arguments:
            if position in self._parameters:
                # validate that the argument can be used for this parameter
                is_valid, v, error_reason = self._parameters[position].validate(value)
                if is_valid:
                    args[self._parameters[position].name] = v
                else:
                    # TODO report/raise exception ??
                    raise Exception("Invalid argument {} used for parameter \"{}\". Reason : {} ".format(value.n3(), self._parameters[position], error_reason))
            else:
                raise Exception("Missing argument in position {} in template {}".format(position, self._name.n3()))
        return args


class MainTemplate(AbstractTemplate):
    """An OTTR template definition, which contains several instances to expand."""

    def __init__(self, name, instances):
        super(MainTemplate, self).__init__(name)
        self._instances = instances

    def __str__(self):
        return self.name + ":: {\n" + ",\n".join(map(lambda i: str(i), self._instances)) + "} ."

    def __repr__(self):
        return self.__str__()

    def expand(self, arguments, all_templates, bnode_suffix=(0, 0),  as_nt=False):
        """Returns a generator that expands the template"""
        for instance in self._instances:
            yield from instance.expand(arguments, all_templates, bnode_suffix=bnode_suffix, as_nt=as_nt)

class NonBaseInstance(AbstractTemplate):
    """
        A non-base OTTR Template instance.
    """

    def __init__(self, name, instance_arguments):
        super(NonBaseInstance, self).__init__(name)
        # store bound & unbound instance arguments separately
        self._bound_arguments = [(x.position, x.value) for x in instance_arguments if x.is_bound]
        self._unbound_arguments = [(x.position, x.value) for x in instance_arguments if not x.is_bound]

    def is_base(self):
        """Returns True if the template is a base template, False otherwise"""
        return False

    def expand(self, arguments, all_templates, bnode_suffix=(0, 0), as_nt=False):
        # increment the bnode unique prefixes, used to unify blank node acrros instance expansions
        bnode_suffix = (bnode_suffix[0], bnode_suffix[1] + 1)
        if self._name in all_templates:
            # fetch template
            template = all_templates[self._name]
            # try to link unbound instance arguments using the given arguments
            args = list(self._bound_arguments)
            for position, value in self._unbound_arguments:
                if value in arguments:
                    args.append((position, arguments[value]))
                else:
                    # TODO raise something ??
                    pass
            # prepare new arguments for recursive template expansion
            new_arguments = dict()
            new_arguments.update(arguments)
            new_arguments.update(template.format_arguments(args))
            # recursively expand the template instance
            yield from template.expand(new_arguments, all_templates, bnode_suffix=bnode_suffix, as_nt=as_nt)
        else:
            raise Exception("Cannot expand the unkown OTTR template '{}'".format(self._name.n3()))



# generator.py
# Author: Thomas MINIER - MIT License 2019
from ottr.parsers import parse_templates, parse_instances
from ottr.tpl import RDF_TEMPLATES, RDFS_TEMPLATES
from rdflib import URIRef, Variable, Literal

class OttrGenerator(object):
    """
        An OttrGenerator can load OTTR templates definitions and expand them to produce RDF triples.
    """

    def __init__(self, load_defaults=True):
        super(OttrGenerator, self).__init__()
        self._templates = dict()
        # counter used for generating instance unique IDs
        self._instance_id = -1
        if load_defaults:
            self.load_templates(RDF_TEMPLATES, format="stottr")
            self.load_templates(RDFS_TEMPLATES, format="stottr")

    def load_templates(self, text, format="stottr"):
        """
            Load a set of OTTR template definitions.

            Arguments:
                - text ``str``: OTTR template definitions in text format.
                - format ``str``: (optional) format of the template definitions. Defaults to sOTTR. Supported formats: sOTTR.

        """
        for template in parse_templates(text, format=format):
            self._templates[template.name] = template

    def instanciate(self, text, format="stottr"):
        """
            Instance a set of OTTR instances.
            Arguments:
                - text ``str``: OTTR instances in text format.
                - format ``str``: (optional) format of the template definitions. Defaults to sOTTR. Supported formats: sOTTR.

            Returns:
                A :class`ottr.generator.OttrInstances`
        """
        # increment the instance ID generator
        self._instance_id += 1
        # parse instances
        instances = parse_instances(text, format=format)
        # create pairs of (instance, related template)
        to_execute = list()
        for instance in instances:
            if instance['name'] in self._templates:
                template = self._templates[instance['name']]
                exec_parameters = template.format_arguments(instance['arguments'])
                to_execute.append((template, exec_parameters))
            else:
                # TODO report error but do not crash??
                pass
        return OttrInstances(self._instance_id, to_execute, self._templates)

class OttrInstances(object):
    """Compiled OTTR instances, ready to be executed to produce RDF triples."""

    def __init__(self, id, to_execute, all_templates):
        super(OttrInstances, self).__init__()
        self._id = id
        self._to_execute = to_execute
        self._all_templates = all_templates

    def execute(self, as_nt=False):
        """
            Execute the instances to produce RDF triples.

            Arguments:
                - as_nt ``bool``: (optional) True if the results should be produced in text RDF format, False if they should be produced in RDFlib format.

            Returns:
                A generator that yields 3-tuples of ``str`` or :class`rdflib.term.Identifier` (depends on the value set for as_nt).
        """
        for template, params in self._to_execute:
            yield from template.expand(params, self._all_templates, bnode_suffix=(self._id, 0), as_nt=as_nt)



# __init__.py
# Author: Thomas MINIER - MIT License 2019
"""
    Manipulate OTTR Reasonable Ontology Templates in Python
"""

from ottr.generator import OttrGenerator

__all__ = [
    'OttrGenerator'
]



# __init__.py
# Author: Thomas MINIER - MIT License 2019
"""
    Lexer and parser utilities for manipulating OTTR template definitions and instances.
"""
from ottr.parsers.stottr.parser import parse_templates_stottr, parse_instances_stottr

__all__ = [
    'parse_templates',
    'parse_instances'
]


def parse_templates(text, format="stottr"):
    """
        Parse a set of stOTTR template definitions.
        Use the format parameter to set the language used to decalred templates (default to pOTTR).
        Formats currently supported: stOTTR
    """
    if format.lower() == 'stottr':
        return parse_templates_stottr(text)
    raise TypeError("Unsupported language '{}'. Only the stOTTR format is currently supported.".format(format))


def parse_instances(text, format="stottr"):
    """
        Parse a set of stOTTR template instances.
        Use the format parameter to set the language used to decalred templates (default to pOTTR).
        Formats currently supported: stOTTR
    """
    if format.lower() == 'stottr':
        return parse_instances_stottr(text)
    raise TypeError("Unsupported language '{}'. Only the stOTTR format is currently supported.".format(format))



# __init__.py
# Author: Thomas MINIER - MIT License 2019
"""
    OTTR template library, available in RDF format at http://tpl.ottr.xyz/
"""

from ottr.tpl.rdf import RDF_TEMPLATES
from ottr.tpl.rdfs import RDFS_TEMPLATES

__all__ = [
    'RDFS_TEMPLATES',
    'RDF_TEMPLATES'
]



# base_templates.py
# Author: Thomas MINIER - MIT License 2019
from ottr.base.template import AbstractTemplate
from ottr.base.argument import URIArgument
from ottr.base.utils import OTTR_TRIPLE_URI
from rdflib.namespace import RDF, RDFS

class OttrTriple(AbstractTemplate):
    """
        The default ottr:Triple base template.
        In pratice, pyOTTR inline each template definition until it contains only instances of base templates (ottr:Triple and ottr:NullableTriple).
    """

    def __init__(self, subject_arg, predicate_arg, object_arg):
        super(OttrTriple, self).__init__(OTTR_TRIPLE_URI)
        self._subject_arg = subject_arg
        self._predicate_arg = predicate_arg
        self._object_arg = object_arg

    def __str__(self):
        return "ottr:Triple ({}, {}, {}) :: BASE .".format(self._subject_arg, self._predicate_arg, self._object_arg)

    def __repr__(self):
        return self.__str__()

    def is_base(self):
        """Returns True if the template is a base template, False otherwise"""
        return True

    def expand(self, arguments, all_templates, bnode_suffix=(0, 0), as_nt=False):
        """Returns a generator that expands the template"""
        yield (
                self._subject_arg.evaluate(bindings=arguments, bnode_suffix=bnode_suffix, as_nt=as_nt),
                self._predicate_arg.evaluate(bindings=arguments, bnode_suffix=bnode_suffix, as_nt=as_nt),
                self._object_arg.evaluate(bindings=arguments, bnode_suffix=bnode_suffix, as_nt=as_nt)
            )



# argument.py
# Author: Thomas MINIER - MIT License 2019
from abc import ABC, abstractmethod
from ottr.base.utils import OTTR
from rdflib import URIRef, BNode

class InstanceArgument(ABC):
    """An instance argument, which corresponds to the parameter of a template."""

    def __init__(self, value, position):
        super(InstanceArgument, self).__init__()
        self._value = value
        self._position = position

    def __str__(self):
        return "({}, pos={})".format(self._value, self._position)

    def __repr__(self):
        return self.__str__()

    @property
    def value(self):
        return self._value

    @property
    def position(self):
        return self._position

    @property
    def is_bound(self):
        """Return True if the argument is bound, False otherwise"""
        return False

    @abstractmethod
    def evaluate(self, bindings=dict(), bnode_suffix=0, as_nt=False):
        """Evaluate the argument using an optional set of bindings"""
        pass


class ConcreteArgument(InstanceArgument):
    """An argument that evaluates to a constant RDF term, i.e., a RDF term."""

    def __init__(self, value, position):
        super(ConcreteArgument, self).__init__(value, position)

    @property
    def is_bound(self):
        return True

    def evaluate(self, bindings=dict(), bnode_suffix=(0,0), as_nt=False):
        term = self._value
        if type(term) == BNode and bnode_suffix is not None:
            term = BNode("{}_{}_{}".format(str(term), bnode_suffix[0], bnode_suffix[1]))
        return term.n3() if as_nt else term

class URIArgument(ConcreteArgument):
    """A ConcreteArgument that evaluates to an URI"""

    def __init__(self, uri, position):
        super(URIArgument, self).__init__(URIRef(uri), position)


class VariableArgument(InstanceArgument):
    """A variable argument, i.e., a SPARQL variable"""

    def __init__(self, value, position):
        super(VariableArgument, self).__init__(value, position)

    def evaluate(self, bindings=dict(), bnode_suffix=(0, 0), as_nt=False):
        if self._value in bindings:
            term = bindings[self._value]
            if type(term) == BNode and bnode_suffix is not None:
                term = BNode("{}_{}_{}".format(str(term), bnode_suffix[0], bnode_suffix[1]))
            return term.n3() if as_nt else term
        return OTTR.none



# rdf.py
# Author: Thomas MINIER - MIT License 2019

RDF_TEMPLATES = """
@prefix o-rdf: <http://tpl.ottr.xyz/rdf/0.1/> .

o-rdf:Type[ ottr:IRI ?entity, ottr:IRI ?type ] :: {
    ottr:Triple(?entity, rdf:type, ?type)
} .
"""



# setup.py
# Author: Thomas MINIER - MIT License 2019
from setuptools import setup, find_packages

__version__ = "0.0.1"

with open('README.rst') as file:
    long_description = file.read()

with open('requirements.txt') as file:
    install_requires = file.read().splitlines()

setup(
    name="ottr",
    version=__version__,
    author="Thomas Minier",
    author_email="thomas.minier@univ-nantes.fr",
    url="https://github.com/Callidon/pyOTTR",
    description="Manipulate OTTR Reasonable Ontology Templates in Python",
    long_description=long_description,
    keywords=["rdf", "owl", "ottr", "template", "ontology", "reusable"],
    license="MIT",
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
    packages=find_packages(exclude=["tests", "tests.*"])
)



# parser.py
# Author: Thomas MINIER - MIT License 2019
from ottr.parsers.stottr.lexer import lex_templates_stottr, lex_instances_stottr
from ottr.base.base_templates import OttrTriple
from ottr.base.template import MainTemplate, NonBaseInstance
from ottr.base.argument import ConcreteArgument, VariableArgument
from ottr.base.utils import OTTR_TRIPLE_URI, OTTR_NONE
from rdflib import Graph, Variable
from rdflib.namespace import RDFS
from rdflib.namespace import NamespaceManager
from rdflib.util import from_n3

# All base templates are registered here,
# as tuples (template constructor, expected nb of arguments)
BASE_TEMPLATES = {
    OTTR_TRIPLE_URI: (OttrTriple, 3)
}

def unify_var(variable, suffix):
    """
        Makes a :class`rdflib.term.Variable` unique by appending a suffix to it.

        Arguments:
            - value :class`rdflib.term.Variable`: Variable to makes unique
            - suffix ``number|str``: suffix to hash & append at the end of the value
        Returns:
            A :class`rdflib.term.Variable`
    """
    return Variable("{}_{}".format(str(variable), suffix))

def get_default_nsm():
    """Get an rdflib NamespaceManager wirh default prefixes configured"""
    nsm = NamespaceManager(Graph())
    # add some prefixes commonly used in pOTTR documents
    nsm.bind('ottr', 'http://ns.ottr.xyz/0.4/')
    nsm.bind('foaf', 'http://xmlns.com/foaf/0.1/')
    nsm.bind('dc', 'http://purl.org/dc/elements/1.1/')
    nsm.bind('owl', 'http://www.w3.org/2002/07/owl#')
    nsm.bind('ax', 'http://tpl.ottr.xyz/owl/axiom/0.1/')
    nsm.bind('rstr', 'http://tpl.ottr.xyz/owl/restriction/0.1/')
    nsm.bind('schema', 'http://schema.org/')
    nsm.bind('schemas', 'https://schema.org/')
    # OTTR base template library
    nsm.bind('o-rdf', 'http://tpl.ottr.xyz/rdf/0.1/')
    nsm.bind('o-rdfs', 'http://tpl.ottr.xyz/rdfs/0.1/')
    return nsm

def parse_term(term, nsm=None):
    """
        Parse a RDF Term from text format to rdflib format.

        Arguments:
            - term ``str``: RDF Term to parse (in text format)
            - nsm :class `rdflib.namespace.NamespaceManager`: (optional) RDFlib namespace manager used to manage prefixes.
    """
    # the special keyword "none" is interpreted as "ottr:None"
    if term == "none":
        return OTTR_NONE
    # rdflib tends to see SPARQL variables as blank nodes, so we need to handle them separately
    if term.startswith('?'):
        return Variable(term[1:])
    return from_n3(term, nsm=nsm)

def parse_instance_arguments(template_id, arguments, nsm=None):
    """
        Parse the arguments of a template instance.

        Arguments:
            - template_id ``int``: ID of the template
            - arguments ``str[]``: instance arguments to parse
            - nsm :class `rdflib.namespace.NamespaceManager`: (optional) RDFlib namespace manager used to manage prefixes.

        Retuns:
         A list of :class`ottr.base.InstanceArgument`
    """
    args = list()
    for ind in range(len(arguments)):
        argument = arguments[ind]
        value = parse_term(argument, nsm=nsm)
        # rewrite variables to make them unique for this scope
        if type(value) is Variable:
            value = unify_var(value, template_id)
            args.append(VariableArgument(value, ind))
        else:
            args.append(ConcreteArgument(value, ind))
    return args

def parse_template_parameter(template_id, param, nsm=None):
    """
        Parse an OTTR template parameter.

        Arguments:
            - template_id ``int``: ID of the template
            - param ``str[]``: parameter to parse
            - nsm :class`rdflib.namespace.NamespaceManager`: (optional) RDFlib namespace manager used to manage prefixes.

        Returns:
         A `dict` representation of the parsed template parameter, with fields "name" (:class`rdflib.term.Identifier`), "type" (:class`rdflib.term.URIRef`), "optional" (``bool``) and "nonblank" (``bool``).
    """
    template_param = dict()
    template_param['name'] = parse_term(param.value, nsm=nsm)
    # rewrite variables to make them unique for this scope
    if type(template_param['name']) == Variable:
        template_param['name'] = unify_var(template_param['name'], template_id)
    template_param['type'] = parse_term(param.type, nsm=nsm) if len(param.type) > 0 else RDFS.Resource
    template_param['optional'] = True if len(param.optional) > 0 else False
    template_param['nonblank'] = True if len(param.nonblank) > 0 else False
    template_param['default'] = parse_term(param.default, nsm=nsm) if len(param.default) > 0 else None
    return template_param

def parse_template_instance(parent_template_id, instance, nsm=None):
    """
        Parse a stOTTR template instance.

        Arguments:
            - parent_template_id ``int``: ID of the template that owns the scope of the instance.
            - instance ``str[]``: instance to parse
            - nsm :class`rdflib.namespace.NamespaceManager`: (optional) RDFlib namespace manager used to manage prefixes.

        Returns:
         A :class`ottr.base.AbstractTemplate`
    """
    template_name = parse_term(instance.name, nsm=nsm)

    # case 1: handle base templates (using a generic method)
    if template_name in BASE_TEMPLATES:
        TemplateConstructor, nb_arguments = BASE_TEMPLATES[template_name]
        if len(instance.arguments) != nb_arguments:
            raise Exception("The {} template takes exactly {} arguments, but {} were provided".format(template_name.n3(), nb_arguments, len(instance.arguments)))
        params = parse_instance_arguments(parent_template_id, instance.arguments, nsm=nsm)
        return TemplateConstructor(*params)

    # case 2: a non-base template instance
    return NonBaseInstance(template_name, parse_instance_arguments(parent_template_id, instance.arguments, nsm=nsm))

def parse_templates_stottr(text):
    """
        Parse a set of stOTTR template definitions and returns the list of all OTTR templates.

        Arguments:
            - text ``str``: stOTTR template definitions in text format.

        Returns:
            A list of :class`ottr.base.MainTemplate` built from the valid sOTTR template definitions provided as input.
    """
    # create a RDFLib NamespaceManager to handle automatic prefix expansion
    nsm = get_default_nsm()

    # run pOTTR lexer
    tree = lex_templates_stottr(text)

    # parse prefixes and register them into the NamespaceManager
    for prefix in tree.prefixes:
        uri = prefix.value
        if uri.startswith('<') and uri.endswith('>'):
            uri = uri[1: -1]
        nsm.bind(prefix.name, uri, replace=True)

    # parse each template definition found
    ottr_templates = list()
    # unique ID sued to make all variables unique to their scope
    template_id = 0
    for template in tree.templates:
        template_name = parse_term(template.name, nsm=nsm)

        # parse instances
        template_instances = [parse_template_instance(template_id, v, nsm=nsm) for v in template.instances]

        # create the OTTR template &
        ottr_template = MainTemplate(template_name, template_instances)
        # parse and register the template's parameters
        for position in range(len(template.parameters.asList())):
            parameter = parse_template_parameter(template_id, template.parameters[position], nsm=nsm)
            ottr_template.add_parameter(parameter['name'], position, param_type=parameter['type'], optional=parameter['optional'], nonblank=parameter['nonblank'], default=parameter['default'])
        # register the new OTTR template
        ottr_templates.append(ottr_template)
        template_id += 1
    return ottr_templates


def parse_instances_stottr(text):
    """
        Parse a set of stOTTR instances and returns them as objects.

        Arguments:
            - text ``str``: stOTTR instances in text format.

        Returns:
            A list of instances built from the valid sOTTR instances provided as input.
    """
    # create a RDFLib NamespaceManager to handle automatic prefix expansion
    nsm = get_default_nsm()

    # run pOTTR lexer
    tree = lex_instances_stottr(text)

    # parse prefixes and register them into the NamespaceManager
    for prefix in tree.prefixes:
        uri = prefix.value
        if uri.startswith('<') and uri.endswith('>'):
            uri = uri[1: -1]
        nsm.bind(prefix.name, uri, replace=True)

    # parse each OTTR instance found
    ottr_instances = list()
    for instance in tree.instances:
        ottr_instance = dict()
        ottr_instance['name'] = parse_term(instance.name, nsm=nsm)
        ottr_instance['arguments'] = list()
        # parse all instance's arguments
        # and save pairs (arguments's position, arguments's RDF value)
        for pos in range(len(instance.arguments)):
            arg = (pos, parse_term(instance.arguments[pos], nsm=nsm))
            ottr_instance['arguments'].append(arg)
        ottr_instances.append(ottr_instance)
    return ottr_instances
